/******************************************************************************
 *
 *  Copyright 2001
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/     

/*******************************************************************************
 *
 * File:         stats.c
 * 
 * What:         statistics diagnostic
 *
 * Description:
 *              This program displays statistical information about crypto and 
 *              key operations on which the BCM device(s) are currently
 *              operating.
 *
 * Revision History:
 *
 *              When       Who   What
 *              12/XX/01   DNA   Created
 *              02/03/02   DNA   Added user help information
 *              02/16/02   DNA   Output Formatting and Help Info
 *
 ******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "ubsio.h"
#include "ubsec_lib.h"


typedef struct ubsec_rate_s {
  double BlocksEncryptedRate;
  double BlocksDecryptedRate;
  double BytesEncryptedRate;
  double BytesDecryptedRate;
  double CryptoFailedRate;
  double IKERate;
  double IKEFailedRate;
  double DHPublicRate;
  double DHSharedRate;
  double RSAPublicRate;
  double RSAPrivateRate;
  double DSASignRate;
  double DSAVerifyRate;
  double DMAErrorRate;
  double DHRate;
  double RSARate;
  double DSARate;
} ubsec_rate_t, *ubsec_rate_pt;


enum {
  false = 0,
  true  = 1
};


enum {
  crypto_mode  = 0x0001,
  key_mode     = 0x0002,
  ssl_mode     = 0x0004,
  loop_mode    = 0x0100,
  verbose_mode = 0x0200,
  clear_mode   = 0x8000
};


void
stats_printHeaderSimple(int control_word) {
  
  if(control_word & key_mode) {
    printf("BCM 58xx Statistics - Public Key Operations\n\n");
    printf(" Dev");
    printf("          IKE");
    printf("        RSA");
    printf("         DH");
    printf("        DSA");
    printf("\n----------------------------------------------------\n");
  } else if(control_word & crypto_mode) {
    printf("BCM 58xx Statistics - Bulk Crypto Operations\n\n");
    printf(" Dev");
    printf("              Enc B");
    printf("               Dec B");
    printf("\n----------------------------------------------\n");
  }
}


void
stats_printHeaderVerbose(int control_word) {

  if(control_word & key_mode) {
    printf("BCM 58xx Statistics - Public Key Operations\n\n");
    printf(" Dev");
    printf("         IKE");
    printf("       DHP");
    printf("       DHS");
    printf("      RPub");
    printf("      RPrv");
    printf("        DS");
    printf("        DV");
    printf("\n------------------------------------------------------------------------------\n");
  } else if(control_word & crypto_mode) {
    printf("BCM 58xx Statistics - Bulk Crypto Operations\n\n");
    printf(" Dev");
    printf("              Enc B");
    printf("               Dec B");
    printf("\n----------------------------------------------\n");
  }
}


void
stats_printRowSimple(ubsec_stats_pt stats, 
		     ubsec_rate_pt rate,
		     int device, 
		     int control_word) {
  
  printf("%4d ", device + 1);
  if(control_word & key_mode) {
    printf("|%11d |", stats->IKECount);
    printf("%9d |", stats->RSAPublicCount + stats->RSAPrivateCount);
    printf("%9d |", stats->DHPublicCount + stats->DHSharedCount);
    printf("%9d |", stats->DSASignCount + stats->DSAVerifyCount);
    printf("\n     ");
    printf("|%11.2f |", rate->IKERate);
    printf("%9.2f |", rate->RSARate);
    printf("%9.2f |", rate->DHRate);
    printf("%9.2f |", rate->DSARate);
    printf("\n\n");
  } else if(control_word & crypto_mode) {
    printf("|%18d |", stats->BytesEncryptedCount);
    printf("%18d |", stats->BytesDecryptedCount);
    printf("\n     ");
    printf("|%18.2f |", rate->BytesEncryptedRate);
    printf("%18.2f |", rate->BytesDecryptedRate);
    printf("\n\n");
  }
}


void
stats_printRowVerbose(ubsec_stats_pt stats, 
		      ubsec_rate_pt rate, 
		      int device, 
		      int control_word) {
  
  printf("%4d ", device + 1);
  if(control_word & key_mode) {
    printf("|%10d |", stats->IKECount);
    printf("%8d |", stats->DHPublicCount);
    printf("%8d |", stats->DHSharedCount);
    printf("%8d |", stats->RSAPublicCount);
    printf("%8d |", stats->RSAPrivateCount);
    printf("%8d |", stats->DSASignCount);
    printf("%8d |", stats->DSAVerifyCount);
    printf("\n     ");
    printf("|%10.2f |", rate->IKERate);
    printf("%8.2f |", rate->DHPublicRate);
    printf("%8.2f |", rate->DHSharedRate);
    printf("%8.2f |", rate->RSAPublicRate);
    printf("%8.2f |", rate->RSAPrivateRate);
    printf("%8.2f |", rate->DSASignRate);
    printf("%8.2f |", rate->DSAVerifyRate);
    printf("\n\n");
  } else if(control_word & crypto_mode) {
    printf("|%18d |", stats->BytesEncryptedCount);
    printf("%18d |", stats->BytesDecryptedCount);
    printf("\n     ");
    printf("|%18.2f |", rate->BytesEncryptedRate);
    printf("%18.2f |", rate->BytesDecryptedRate);
    printf("\n\n");
  }
}


void
stats_printTableSimple(ubsec_stats_t stats[],
		       ubsec_rate_t rate[],
		       int num_devices, 
		       int current_device, 
		       int control_word) {
  
  int i;
  
  if(control_word & (clear_mode | loop_mode))
    system("clear");
  stats_printHeaderSimple(control_word);
  
  if(current_device == num_devices)
    for(i = 0; i < num_devices; i++)
      stats_printRowSimple(&(stats[i]), 
			   &(rate[i]), 
			   i, 
			   control_word);
  else
    stats_printRowSimple(&(stats[current_device]), 
			 &(rate[current_device]), 
			 current_device, 
			 control_word);
}


void
stats_printTableVerbose(ubsec_stats_t stats[],
			ubsec_rate_t rate[],
			int num_devices, 
			int current_device, 
			int control_word) {
  
  int i;
  
  if(control_word & (clear_mode | loop_mode))
    system("clear");
  stats_printHeaderVerbose(control_word);
  
  if(current_device == num_devices)
    for(i = 0; i < num_devices; i++)
      stats_printRowVerbose(&(stats[i]), 
			    &(rate[i]), 
			    i, 
			    control_word);
  else
    stats_printRowVerbose(&(stats[current_device]), 
			  &(rate[current_device]), 
			  current_device, 
			  control_word);
}


void
stats_printTable(ubsec_stats_t stats[],
		 ubsec_rate_t rate[],
		 int num_devices, 
		 int current_device, 
		 int control_word) {
  
  if(control_word & verbose_mode)
    stats_printTableVerbose(stats, rate, num_devices, current_device, control_word);
  else
    stats_printTableSimple(stats, rate, num_devices, current_device, control_word);
}


void
stats_rateCalc(ubsec_stats_pt stats,
	       ubsec_stats_pt prev_stats,
	       int interval,
	       ubsec_rate_pt rate) {
  
  rate->BlocksEncryptedRate = 
    (stats->BlocksEncryptedCount - prev_stats->BlocksEncryptedCount) / (double) interval;
  rate->BlocksDecryptedRate = 
    (stats->BlocksEncryptedCount - prev_stats->BlocksEncryptedCount) / (double) interval;
  rate->BytesEncryptedRate = 
    (stats->BytesEncryptedCount - prev_stats->BytesEncryptedCount) / (double) interval;
  rate->BytesDecryptedRate = 
    (stats->BytesDecryptedCount - prev_stats->BytesDecryptedCount) / (double) interval;
  rate->CryptoFailedRate = 
    (stats->CryptoFailedCount - prev_stats->CryptoFailedCount) / (double) interval;
  rate->IKERate = 
    (stats->IKECount - prev_stats->IKECount) / (double) interval;
  rate->IKEFailedRate = 
    (stats->IKEFailedCount - prev_stats->IKEFailedCount) / (double) interval;
  rate->DHPublicRate = 
    (stats->DHPublicCount - prev_stats->DHPublicCount) / (double) interval;
  rate->DHSharedRate = 
    (stats->DHSharedCount - prev_stats->DHSharedCount) / (double) interval;
  rate->RSAPublicRate = 
    (stats->RSAPublicCount - prev_stats->RSAPublicCount) / (double) interval;
  rate->RSAPrivateRate = 
    (stats->RSAPrivateCount - prev_stats->RSAPrivateCount) / (double) interval;
  rate->DSASignRate = 
    (stats->DSASignCount - prev_stats->DSASignCount) / (double) interval;
  rate->DSAVerifyRate = 
    (stats->DSAVerifyCount - prev_stats->DSAVerifyCount) / (double) interval;
  rate->DMAErrorRate = 
    (stats->DMAErrorCount - prev_stats->DMAErrorCount) / (double) interval;

  rate->DHRate = 
    (stats->DHPublicCount + stats->DHSharedCount - 
     prev_stats->DHPublicCount - prev_stats->DHSharedCount) / (double) interval;
  rate->RSARate =
    (stats->RSAPublicCount + stats->RSAPrivateCount - 
     prev_stats->RSAPublicCount - prev_stats->RSAPrivateCount) / (double) interval;
  rate->DSARate =
    (stats->DSASignCount + stats->DSAVerifyCount -
     prev_stats->DSASignCount - prev_stats->DSAVerifyCount) / (double) interval;
}


void
stats_computeRate(ubsec_stats_t stats[],
		  ubsec_stats_t prev_stats[],
		  int interval,
		  int num_devices,
		  int current_device,
		  ubsec_rate_t rate[]) {
  
  static int init_flag = true;
  int i;
  
  if(init_flag == true) {
    init_flag = false;
    bzero(rate, sizeof(ubsec_rate_t));
  } else {
    if(current_device == num_devices)
      for(i = 0; i < num_devices; i += 1)
	stats_rateCalc(&(stats[i]), 
		       &(prev_stats[i]), 
		       interval, 
		       &(rate[i]));
    else
      stats_rateCalc(&(stats[current_device]), 
		     &(prev_stats[current_device]), 
		     interval, 
		     &(rate[current_device]));
  }
}


int
stats_numDevices(void) {
  
  int ubsec_fd;
  int numDevices;
  
  if ((ubsec_fd = ubsec_open(UBSEC_KEY_DEVICE)) <= 0) {
    printf("Failed to open %s\n.", UBSEC_KEY_DEVICE);
    return(-1);
  }
  
  ubsec_numDevices_ioctl(ubsec_fd, &numDevices);

  ubsec_close(ubsec_fd);

  return numDevices;
}


void
stats_helpInfo(void) {

  printf("\nBCM Statistics Diagnostic\n");
  printf("Usage Information:\n\n");
  printf("\t-l\t\tLoop mode. The diagnostic continues to collect and\n\t\t\tdisplay statistics until cancelled by the user.\n\n");
  printf("\t-i [num]\tSet interval. Adjust the interval used in loop mode to\n\t\t\tsample and display new statistics.\n\n");
  printf("\t-v\t\tVerbose mode. Show more detailed statistics information\n\t\t\tincluding rate information.\n\n");
  printf("\t-s [num]\tShow statistics for specified device only. Default is\n\t\t\tall devices.\n\n");
  printf("\t-c\t\tShow statistics for crypto operations. Default is key\n\t\t\toperations only.\n\n");
  printf("\t-n\t\tAvoid clearing the screen when not in loop mode.\n\n");
  printf("\t-h\t\tDisplay usage information.\n\n");
}


int
main(int argc, char *argv[]) {
  
  int num_devices;
  int current_device;
  int interval;
  int control_word;
  int c;
  int i;
  int status;
  int ubsec_fd;
  ubsec_stats_t stats[/*MAX_DEVICES*/128];
  ubsec_stats_t prev_stats[/*MAX_DEVICES*/128];
  ubsec_rate_t rate[/*MAX_DEVICES*/128];
  
  /*************************************** Initialization */
  
  num_devices = stats_numDevices();
  current_device = num_devices;
  
  interval   = 3;
  control_word  = 0;
  control_word |= key_mode;
  control_word |= clear_mode;
  
  bzero(prev_stats, sizeof(ubsec_stats_t));

  while ((c = getopt(argc, argv, "LlVvS:s:I:i:CcDdKknHh?")) != -1) {
    switch (c) {
    case 'L':
    case 'l':
      control_word |= loop_mode;
      break;
    case 'I': // second arg to looping mode...
    case 'i':
      interval = atoi(optarg);
    case 'V':
    case 'v':
      control_word |= verbose_mode;
      break;
    case 'S':
    case 's':
      current_device = atoi(optarg);
      if (current_device == 0)
	      if (optarg[0] != '0')
	      {
		      printf("Invalid card number.\n");
		      goto failed_arg;
		}
      break;
    case 'C':
    case 'c':
      control_word &= (control_word & ~key_mode);
      control_word |= crypto_mode;
      break;
    case 'K':
    case 'k':
      control_word &= (control_word & ~crypto_mode);
      control_word |= key_mode;
      break;
    case 'n':
      control_word &= (control_word & ~clear_mode);
      break;
    case 'H':
    case 'h':
    case '?':
      stats_helpInfo();
      return;
      break;
    default:
      break;
    }
  }

  /*********************************** Collect Statistics */

  if(control_word & key_mode) {
    if((ubsec_fd = ubsec_open(UBSEC_KEY_DEVICE)) <= 0) {
      status = ubsec_fd;
      printf("Failed to open %s\n.", UBSEC_KEY_DEVICE);
      return -1;
    }
  } else if(control_word & crypto_mode) {
    if((ubsec_fd = ubsec_open(UBSEC_CRYPTO_DEVICE)) <= 0) {
      status = ubsec_fd;
      printf("Failed to open %s\n.", UBSEC_CRYPTO_DEVICE);
      return -1;
    }
  }

  if (current_device > num_devices)
  {
	if (num_devices > 1)
        	printf("Invalid card number specified. Card number should be between 1 and %d.\n", num_devices);
	else
        	printf("Invalid card number specified. Card number should %d.\n", num_devices);

	goto failed_exit;
  }

  do {
    
    if(current_device == num_devices) {
      for(i = 0; i < num_devices; i += 1)
	status = ubsec_stats_ioctl(ubsec_fd, i, &(stats[i]));
    } else {

      status = ubsec_stats_ioctl(ubsec_fd, current_device, &(stats[current_device]));
    }

    if (status < 0)
	{
		printf("ubsec_stats_ioctl failed (%d)",status);
		goto failed_exit;
	}
    
    stats_computeRate(stats, prev_stats, interval, num_devices, current_device, rate);
    stats_printTable(stats, rate, num_devices, current_device, control_word);
    
    if(control_word & loop_mode)
      sleep(interval);
    
    for(i = 0; i < /*MAX_DEVICES*/128; i++) /* hang onto last interval's stats to compute rate */
      prev_stats[i] = stats[i];
    
  } while(control_word & loop_mode);

failed_exit:
  
  ubsec_close(ubsec_fd);
failed_arg:
  ;
}
