/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/
/* 
 * Broadcom Corporation uBSec SDK 
 */
/*
 * ubsec_key_ioctl.c:  Key setup functions, ioctl versions.
 */
/*
 * Revision History:
 *
 * 05/24/00 JJT Created.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include "ubsec.h"
#include "ubsio.h"
#include "ubsec_lib.h"

unsigned long user_timing;

void
ubsec_ioctl_time(unsigned long *time_us) {
	*time_us = user_timing;
}

/*
 * Diffie-Hellman routines.
 *
 * All buffer lengths are given in bits. All input is assumed 'little endian' ordered,
 * first byte contains the LSB.
 */

/*
 * diffie_hellman_generate:  compute  y = g**x | m.
 *
 * Private value x, and public value y are output; generator g and modulus m are input only.
 * UserX contains an optional user supplied private value, which either must be supplied or
 * random_bits must be greater than zero. Results are copied into caller supplied buffers.
 */
int
diffie_hellman_generate_ioctl(int fd, unsigned char *x, int *x_len, unsigned char *y, 
	int *y_len, unsigned char *g, int g_len, unsigned char *m, int m_len,
	unsigned char *userX, int userX_len, int random_bits)
{
	ubsec_key_io_t		cmd;
	ubsec_DH_Params_t	*params;
	int			status = 0;

	if ((*y_len < m_len) || (*x_len < m_len)) {
		return	UBSEC_STATUS_INVALID_PARAMETER;
	}

	bzero(&cmd, sizeof(cmd));
	cmd.command = UBSEC_DH_PUBLIC;
	params = &(cmd.key.DHParams);

	params->N.KeyLength = m_len;
	params->N.KeyValue = m;

	/*
	 * Put in the generator.
	 */
	params->G.KeyLength = g_len;
	params->G.KeyValue = g;

	/*
	 * Use user secret if one is supplied, otherwise generate.
	 */
	if (userX && userX_len){
		params->UserX.KeyLength = userX_len;
		params->UserX.KeyValue = userX;
	} else {
		params->RNGEnable = 1;
		params->RandomKeyLen = random_bits;
	}

	/*
	 * Public (y) and secret (x) values will come back.
	 */
	params->Y.KeyValue = y;
	params->Y.KeyLength = *y_len;

	params->X.KeyValue = x;
	params->X.KeyLength = *x_len;

	/*
	 * Call chip.
	 */
	status = ubsec_ioctl (fd, (unsigned char *)&cmd, UBSEC_KEY_SETUP_FUNC);

	/*
	 * Copy back results.
	 */
	if (status == UBSEC_STATUS_SUCCESS) {
		*y_len = params->Y.KeyLength;
		*x_len = params->X.KeyLength;
	}

	user_timing = cmd.time_us;

#ifdef TIMING
	if( user_timing != 0 )
		printf("Speed is %f mbps -- %luus\n", 
		(8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */

	return status;
}

/*
 * diffie_hellman_agree:  compute  k = y**x | m.
 *
 * User private value x, received Y-value y, and modulus m  are input.
 * Shared secret key is output to callers buffer.
 */
int
diffie_hellman_agree_ioctl(int fd, unsigned char *x, int x_len, unsigned char *y, 
	int y_len, unsigned char *m, int m_len, unsigned char *k, int *k_len)
{
	ubsec_key_io_t		cmd;
	ubsec_DH_Params_t	*params;
	int			status = 0;

	if (*k_len < m_len) {
		return	UBSEC_STATUS_INVALID_PARAMETER;
	}

	bzero(&cmd, sizeof(cmd));
	cmd.command = UBSEC_DH_SHARED;
	params = &(cmd.key.DHParams);

	params->N.KeyLength = m_len;
	params->N.KeyValue = m;


	params->X.KeyLength = x_len;
	params->X.KeyValue = x;

	params->Y.KeyLength = y_len;
	params->Y.KeyValue = y;

	params->K.KeyValue = k;
	params->K.KeyLength = *k_len;

	/*
	 * Call chip.
	 */
	status = ubsec_ioctl (fd, (unsigned char *)&cmd, UBSEC_KEY_SETUP_FUNC);

	/*
	 * Copy back results.
	 */
	if (status == UBSEC_STATUS_SUCCESS) {
		*k_len = params->K.KeyLength;
	}

	user_timing = cmd.time_us;

#ifdef TIMING
	if( user_timing != 0 )
		printf("Speed is %f mbps -- %luus\n", 
		(8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */

	return status;
}

/*
 * RSA public key routines.
 *
 * All buffer lengths are supplied in bits.
 *
 */

/*
 * rsa_mod_exp:  compute modular exponential y = x**e | m.
 */
int
rsa_mod_exp_ioctl (int fd, unsigned char *x, int x_len, unsigned char *m, int m_len,
		unsigned char *e, int e_len, unsigned char *y, int *y_len)
{
	ubsec_key_io_t		cmd;
	ubsec_RSA_Params_t	*params;
	int			status;

	if (*y_len < m_len) {
		return	UBSEC_STATUS_INVALID_PARAMETER;
	}

	bzero(&cmd, sizeof(cmd));
	cmd.command = UBSEC_RSA_PUBLIC;
	params = &(cmd.key.RSAParams);

	params->ModN.KeyLength = m_len;
	params->ModN.KeyValue = m;

	params->ExpE.KeyLength = e_len;
	params->ExpE.KeyValue = e;

	params->InputKeyInfo.KeyLength = x_len;
	params->InputKeyInfo.KeyValue = x;

	/*
	 *  result comes back in y.
	 */
	params->OutputKeyInfo.KeyValue = y;
	params->OutputKeyInfo.KeyLength = *y_len;

	/*
	 * Call chip.
	 */
	status = ubsec_ioctl (fd, (unsigned char *)&cmd, UBSEC_KEY_SETUP_FUNC);

	/*
	 * Copy back results.
	 */
	if (status == UBSEC_STATUS_SUCCESS) {
		*y_len = params->OutputKeyInfo.KeyLength;
	}

	user_timing = cmd.time_us;

#ifdef TIMING
	if( user_timing != 0 )
		printf("Speed is %f mbps -- %luus\n", 
		(8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */

	return status;
}

/*
 * mod_exp_crt:  compute modular exponential using Chinese remainder theorem:
 *			 y1 = (x|p)**edp | p
 *			 y2 = (x|q)**edq | q
 *			 y = y2 + (qinv * (y1 - y2) | p) * q
 */
int
rsa_mod_exp_crt_ioctl (int fd, unsigned char *x, int x_len, unsigned char *qinv, 
	int qinv_len, unsigned char *edq, int edq_len, unsigned char *q, int q_len,
	unsigned char *edp, int edp_len, unsigned char *p, int p_len,
	unsigned char *y, int *y_len)
{
	ubsec_key_io_t		cmd;
	ubsec_RSA_Params_t	*params;
	int			status;

	if (*y_len < q_len) {
		return	UBSEC_STATUS_INVALID_PARAMETER;
	}

	bzero(&cmd, sizeof(cmd));
	cmd.command = UBSEC_RSA_PRIVATE;
	params = &(cmd.key.RSAParams);

	params->PrimeP.KeyLength = p_len;
	params->PrimeP.KeyValue = p;

	params->PrimeQ.KeyLength = q_len;
	params->PrimeQ.KeyValue = q;

	params->PrimeEdp.KeyLength = edp_len;
	params->PrimeEdp.KeyValue = edp;

	params->PrimeEdq.KeyLength = edq_len;
	params->PrimeEdq.KeyValue = edq;

	params->Pinv.KeyLength = qinv_len;
	params->Pinv.KeyValue = qinv;

	params->InputKeyInfo.KeyLength = x_len;
	params->InputKeyInfo.KeyValue = x;

	params->OutputKeyInfo.KeyValue = y;
	params->OutputKeyInfo.KeyLength = *y_len;

	/*
	 * Call chip.
	 */
	status = ubsec_ioctl (fd, (unsigned char *)&cmd, UBSEC_KEY_SETUP_FUNC);

	/*
	 * Copy back results.
	 */
	if (status == UBSEC_STATUS_SUCCESS) {
		*y_len = params->OutputKeyInfo.KeyLength;
	}

	user_timing = cmd.time_us;

#ifdef TIMING
	if( user_timing != 0 )
		printf("Speed is %f mbps -- %luus\n", 
		(8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */

	return status;
}

/*
 * DSA public key routines.
 *
 */

/*
 * dsa_sign. Output is signature < r, s >
 *
 */
int
dsa_sign_ioctl (int fd, int hash, unsigned char *data, int data_len, 
		unsigned char *random, int random_len, 
		unsigned char *p, int p_len, unsigned char *q, int q_len,
		unsigned char *g, int g_len, unsigned char *key, int key_len,
		unsigned char *r, int *r_len, unsigned char *s, int *s_len)
{
	ubsec_key_io_t		cmd;
	ubsec_DSA_Params_t	*params;
	int			status;
	ubsec_FragmentInfo_t	InputData;

	if ((*r_len < q_len) || (*s_len < q_len)) {
		return	UBSEC_STATUS_INVALID_PARAMETER;
	}

	bzero(&cmd, sizeof(cmd));
	cmd.command = UBSEC_DSA_SIGN;
	params = &(cmd.key.DSAParams);

	params->ModQ.KeyLength = q_len;
	params->ModQ.KeyValue = q;

	params->ModP.KeyLength = p_len;
	params->ModP.KeyValue = p;

	params->BaseG.KeyLength = g_len;
	params->BaseG.KeyValue = g;

	params->Key.KeyLength = key_len;
	params->Key.KeyValue = key;

	params->InputFragments = &InputData;
	params->NumInputFragments=1;
	params->InputFragments->FragmentLength = (data_len + 7)/8;
	params->InputFragments->FragmentAddress = (ubsec_MemAddress_t)data;

	if (random_len && random) {
		params->Random.KeyLength = random_len;
		params->Random.KeyValue = random;
	} else {
		params->RandomLen = SHA_BIT_SIZE;
		params->RNGEnable = 1;
	}
	
	if (hash) {
		params->HashEnable = 1;
	}

	params->SigR.KeyLength = *r_len;
	params->SigR.KeyValue = r;

	params->SigS.KeyLength = *s_len;
	params->SigS.KeyValue = s;

	/*
	 * Call chip.
	 */
	status = ubsec_ioctl (fd, (unsigned char *)&cmd, UBSEC_KEY_SETUP_FUNC);

	if (status == UBSEC_STATUS_SUCCESS) {
		*r_len = params->SigR.KeyLength;
		*s_len = params->SigS.KeyLength;
	}

	user_timing = cmd.time_us;

#ifdef TIMING
	if( user_timing != 0 )
		printf("Speed is %f mbps -- %luus\n", 
		(8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */


	return status;
}

/*
 * dsa_verify. Output is verification value v.
 */
int
dsa_verify_ioctl (int fd, int hash, unsigned char *data, int data_len,
		unsigned char *p, int p_len, unsigned char *q, int q_len,
		unsigned char *g, int g_len, unsigned char *key, int key_len,
		unsigned char *r, int r_len, unsigned char *s, int s_len,
		unsigned char *v, int *v_len)
{
	ubsec_key_io_t		cmd;
	ubsec_DSA_Params_t	*params;
	int			out_len,
				status;
	ubsec_FragmentInfo_t	InputData;

	if (*v_len < q_len) {
		return	UBSEC_STATUS_INVALID_PARAMETER;
	}

	bzero(&cmd, sizeof(cmd));
	cmd.command = UBSEC_DSA_VERIFY;
	params = &(cmd.key.DSAParams);

	params->ModQ.KeyLength = q_len;
	params->ModQ.KeyValue = q;

	params->ModP.KeyLength = p_len;
	params->ModP.KeyValue = p;

	params->BaseG.KeyLength = g_len;
	params->BaseG.KeyValue = g;

	params->Key.KeyLength = key_len;
	params->Key.KeyValue = key;

	params->InputFragments = &InputData;
	params->NumInputFragments=1;
	params->InputFragments->FragmentLength = (data_len + 7)/8;
	params->InputFragments->FragmentAddress = (ubsec_MemAddress_t)data;

	params->SigR.KeyLength = r_len;
	params->SigR.KeyValue = r;

	params->SigS.KeyLength = s_len;
	params->SigS.KeyValue = s;

	if (hash) {
		params->HashEnable = 1;
	}

	out_len = ((q_len + 31)/32) * sizeof(long);

	params->V.KeyValue = v;
	params->V.KeyLength = *v_len;

	/*
	 * Call chip.
	 */
	status = ubsec_ioctl (fd, (unsigned char *)&cmd, UBSEC_KEY_SETUP_FUNC);

	/*
	 * Copy back verification value.
	 */
	if (status == UBSEC_STATUS_SUCCESS) {
		*v_len = params->V.KeyLength;
	}

	user_timing = cmd.time_us;

#ifdef TIMING
	if( user_timing != 0 )
		printf("Speed is %f mbps -- %luus\n", 
		(8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */

	return status;
}
