/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/

/* 
 * Broadcom Corporation uBSec SDK 
 */

/*
 * ubsec_lib.h:  Header file for library APIl
 */

/*
 * Revision History:
 *
 * When      Who  What
 * 05/24/00  JJT  Created.
 * 09/20/00  DNA  Added SSL Ioctl Function Prototypes
 *
 */

/*
 * All buffer lengths are supplied in bits.
 *
 */

#ifndef _UBSEC_LIB_H_
#define _UBSEC_LIB_H_

/* UBSEC_LIB Version  */

#define UBSEC_LIB_VERSION_MAJOR 0x1
#define UBSEC_LIB_VERSION_MINOR 0x1
#define UBSEC_LIB_VERSION_REV 	' '


#define MAX_PUBLIC_KEY_BITS (2048)
#define MAX_RNG_BITS (1024)
#define MAX_MATH_BITS (2048)

#define MAX_PUBLIC_KEY_BYTES (ROUNDUP_TO_32_BIT(MAX_PUBLIC_KEY_BITS)/8)
#define MAX_RNG_BYTES (ROUNDUP_TO_32_BIT(MAX_RNG_BITS)/8)
#define MAX_MATH_BYTES (ROUNDUP_TO_32_BIT(MAX_MATH_BITS)/8)

#define BITSTOBYTES(bitsize) (ROUNDUP_TO_32_BIT(bitsize)/8)

#define SHA_BIT_SIZE  (160)

#ifdef CRYPTONET_DEVICE
#define UBSEC_SSL_DEVICE 	CRYPTONET_DEVICE 
#define UBSEC_KEY_DEVICE 	CRYPTONET_DEVICE
#define UBSEC_CRYPTO_DEVICE	CRYPTONET_DEVICE 
#else
#define UBSEC_SSL_DEVICE "/dev/cryptonet"
#define UBSEC_KEY_DEVICE "/dev/cryptonet"
#define UBSEC_CRYPTO_DEVICE "/dev/cryptonet"
#endif

#ifdef BCM_OEM_1
#include "bcm_oem1_device.h"
#endif

typedef void ubsec_callback(ubsec_rw_pt rw_cmd, void *user_arg);

/*
 * Raw I/O routines. Normally, only ubsec_open and ubsec_close would be called
 * by applications.
 */
int ubsec_open(char *device);
int ubsec_close(int fd);
int ubsec_ioctl(int ubsec_fd, unsigned char *buf, long cmd);

void ubsec_ioctl_time(unsigned long *time_us);

/*
 * General utility routines.
 */
int ubsec_bytes_to_bits(unsigned char *n, int bytes);
int ubsec_bits_to_bytes(int bits);

int ubsec_keysetup_callback(ubsec_rw_pt rw_cmd, unsigned long *time_us);
int ubsec_keysetup_completion(int fd, unsigned long *time_us);

/*
 * ubsec_chipinfo_ioctl: fills ubsec_chipinfo_io_t structure with device/driver info.
 */ 
int ubsec_chipinfo_ioctl(int fd, ubsec_chipinfo_io_t *chipinfo_structure);


/*
 *  Function: ubsec_rsakeygen
 *
 *  Description: This function takes as its inputs a public
 *               exponent and the lengths of the private
 *               values p and q. From these, this function
 *               computes all of the public and private
 *               keys necessary for RSA Public and Private
 *               CRT operations.
 *
 *  Return Values:
 *                 == 0  =>  OK
 *                 != 0  =>  Error
 */

int
ubsec_rsakeygen(
		int fd,
		int modulus_bits,
		unsigned char *  e,    int *  e_bits,
		unsigned char *  p,    int *  p_bits,
		unsigned char *  q,    int *  q_bits,
		unsigned char *  n,    int *  n_bits,
		unsigned char *  d,    int *  d_bits,
		unsigned char *  dp,   int *  dp_bits,
		unsigned char *  dq,   int *  dq_bits,
		unsigned char *  pinv, int *  pinv_bits);


/*
 * Ioctl based routines, make synchronous ioctl calls.
 */
int diffie_hellman_generate_ioctl(int fd, unsigned char *x, int *x_len, unsigned char *y, 
	int *y_len, unsigned char *g, int g_len, unsigned char *m, int m_len,
	unsigned char *userX, int userX_len, int random_bits);

int diffie_hellman_agree_ioctl(int fd, unsigned char *x, int x_len, unsigned char *y, 
	int y_len, unsigned char *m, int m_len, unsigned char *k, int *k_len);

int rsa_mod_exp_ioctl (int fd, unsigned char *x, int x_len, unsigned char *m, int m_len,
	unsigned char *e, int e_len, unsigned char *y, int *y_len);

int rsa_mod_exp_crt_ioctl (int fd, unsigned char *x, int x_len, unsigned char *qinv, 
	int qinv_len, unsigned char *edq, int edq_len, unsigned char *q, int q_len,
	unsigned char *edp, int edp_len, unsigned char *p, int p_len,
	unsigned char *y, int *y_len);

int dsa_sign_ioctl (int fd, int hash, unsigned char *data, int data_len, 
	unsigned char *random, int random_len, 
	unsigned char *p, int p_len, unsigned char *q, int q_len,
	unsigned char *g, int g_len, unsigned char *key, int key_len,
	unsigned char *r, int *r_len, unsigned char *s, int *s_len);

int dsa_verify_ioctl (int fd, int hash, unsigned char *data, int data_len,
	unsigned char *p, int p_len, unsigned char *q, int q_len,
	unsigned char *g, int g_len, unsigned char *key, int key_len,
	unsigned char *r, int r_len, unsigned char *s, int s_len,
	unsigned char *v, int *v_len);

/* Math acceleration routines. */
int math_accelerate_write(int fd, ubsec_MathCommand_t command,
	unsigned char *ModN, int *ModN_len, unsigned char *unused, int *unused_len, 
	unsigned char *ParamA, int *ParamA_len, unsigned char *ParamB, int *ParamB_len,
	unsigned char *Result, int *Result_len,
	ubsec_callback fn, void *arg);

int math_accelerate_ioctl(int fd, ubsec_MathCommand_t command,
	unsigned char *ModN, int *ModN_len, unsigned char *unused, int *unused_len, 
	unsigned char *ParamA, int *ParamA_len, unsigned char *ParamB, int *ParamB_len,
	unsigned char *Result, int *Result_len);

int math_accelerate_dblmodexp_ioctl(int fd, ubsec_MathCommand_t command,
	unsigned char *ModN0, int *ModN0_len, unsigned char *ModN1, int *ModN1_len, 
	unsigned char *ParamA0, int *ParamA0_len, unsigned char *ParamA1, int *ParamA1_len, 
	unsigned char *ParamE0, int *ParamE0_len, unsigned char *ParamE1, int *ParamE1_len,
	unsigned char *Result0, int *Result0_len, unsigned char *Result1, int *Result1_len);

int
ubsec_modrem(
	     int fd,
	     unsigned char *a, int *a_bits,
	     unsigned char *n, int *n_bits,
	     unsigned char *c, int *c_bits
	     );

int
ubsec_modinv(
	     int fd,
	     unsigned char *n, int *n_bits,
	     unsigned char *b, int *b_bits,
	     unsigned char *t, int *t_bits
	     );

/* Random number generation routines. */
int rng_ioctl(int fd, ubsec_RNGCommand_t command,
	unsigned char *Result, int *Result_len);

int
ubsec_primerng(
	       int                 fd,
	       ubsec_RNGCommand_t  rng_command,
	       int                 length,
	       unsigned char      *result,
	       int                *result_bits);

/* Symmetric encryption routines */

typedef struct ubsec_crypto_context_s {
	unsigned int	flags;
	unsigned char	crypto[MAX_CRYPT_KEY_LENGTH];
	unsigned char	auth[MAC_KEY_LENGTH];
} ubsec_crypto_context_t, *ubsec_crypto_context_p;

/*
 * Initialize crypto context.
 */
int ubsec_crypto_init(unsigned char *k1, unsigned char *k2, unsigned char *k3,
	unsigned char *auth_key, int encr_alg, int auth_alg, ubsec_crypto_context_p ctx);

int ubsec_crypto_done(ubsec_crypto_context_p ctx);

/*
 * Pass a buffer to encrypt/decrypt/authenticate.
 */
int ubsec_crypto_data_ioctl(int fd, int cmd, ubsec_crypto_context_p ctx, 
	unsigned char *data_in, unsigned char *iv, unsigned short in_len,
	unsigned short crypto_offset, unsigned char *data_out, unsigned short out_len,
	unsigned long *optime);

/*
 * Ubecom Routines
 */

/* SSL Routines */

int
ubsec_ssl_mac_ioctl(
		    int              fd,
		    unsigned char *  data_in,      /* data to authenticate */
		    unsigned int     data_in_len,  /* length of data to auth */
		    unsigned char *  auth_key,     /* secret key */
		    unsigned int     auth_key_len, /* length of secret key */
		    unsigned char *  seq_num,      /* sequence number */
		    unsigned char    type,         /* content type */
		    unsigned short   data_length,  /* data length */
		    unsigned int     auth_alg,     /* hash agthm: sha or md5 */
		    unsigned char *  data_out,     /* buffer for output data */
		    unsigned int     data_out_len  /* length of output buffer */
		    );

int
ubsec_ssl_3des_ioctl(
		      int fd,
		      unsigned char *data_in,      /* input data buffer -- to be encrypted/decrypted */
		      unsigned int data_in_len,    /* length input data buffer */
		      unsigned char *key1,         /* first DES key */
		      unsigned char *key2,         /* second DES key */
		      unsigned char *key3,         /* third DES key */
		      unsigned char *iv,           /* initial vector */
		      unsigned int cmd,            /* encrypt or decrypt */
		      unsigned char *data_out,     /* output data buffer */
		      unsigned int data_out_len    /* length of output data buffer */
		      );

int
ubsec_ssl_arc4_init_ioctl(
			  int fd,
			  unsigned char *  key,        /* key data */
			  unsigned int     key_len,    /* length in bytes of key */
			  unsigned char    index_i,    /* initial value of index i */
			  unsigned char    index_j,    /* initial value of index j */
			  unsigned char *  key_buf,    /* output key/index buffer */
			  unsigned int     key_buf_len /* length of output key/index buffer in bytes */
			  );

int
ubsec_ssl_arc4_process_ioctl(
			     int fd,
			     unsigned char *data_in,      /* input data buffer */
			     unsigned int data_in_len,    /* length of input buffer */
			     unsigned char *key_state,    /* key or state */
			     unsigned int key_state_flag, /* start from key or state */
			     unsigned char *data_out,     /* output data buffer */
			     unsigned int data_out_len,   /* length of output buffer */
			     unsigned char *state_out     /* state upon completion */
			     );

/* TLS Routines */

int
ubsec_tls_mac_ioctl(
		    int              fd,
		    unsigned char *  data_in,      /* data to authenticate */
		    unsigned int     data_in_len,  /* length of data to auth */
		    unsigned char *  auth_key,     /* MAC Secret Key */
		    unsigned int     auth_key_len, /* MAC Secret Key Length in Bytes */
		    unsigned char *  seq_num,      /* sequence number */
		    unsigned long    version,      /* TLS Version */
		    unsigned char    type,         /* TLS Content Type */
		    unsigned long    datalength,   /* data length */
		    unsigned int     auth_alg,     /* hash agthm: sha or md5 */
		    unsigned char *  data_out,     /* buffer for output data */
		    unsigned int     data_out_len  /* length of output buffer */
		    );

/* Hash Routines */
int
ubsec_hash_ioctl(
		 int fd,
		 unsigned char *data_in,          /* input data buffer */
		 unsigned int data_in_len,        /* length of input data buffer */
		 int auth_alg,                    /* sha or md5 hash alg */
		 unsigned char *data_out,         /* output data buffer */
		 unsigned int data_out_len        /* length of output data buffer */
		 );

/*********************************************************** Statistics */
typedef struct ubsec_stats_s {
  int BlocksEncryptedCount;
  int BlocksDecryptedCount;
  int BytesEncryptedCount;
  int BytesDecryptedCount;
  int CryptoFailedCount;
  int IKECount;
  int IKEFailedCount;
  int DHPublicCount;
  int DHSharedCount;
  int RSAPublicCount;
  int RSAPrivateCount;
  int DSASignCount;
  int DSAVerifyCount;
  int DMAErrorCount;
} ubsec_stats_t, *ubsec_stats_pt;
 
 
int
ubsec_stats_ioctl(int ubsec_fd,
                  int device_num,
                  ubsec_stats_pt stats);

#endif /* _UBSEC_LIB_H_ */




