/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/

/* 
 *  Broadcom Corporation uBSec SDK 
 */

/*
 *  File: ubsec_math.c
 *
 *  Desc: Math functions
 *
 */

/*
 *  Revision History:
 *
 *  05/24/00 JJT Created.
 *  12/06/00 DNA Added separate function for ModRem
 *  12/06/00 DNA Added separate functions for ModInv
 *
 */

#include <stdio.h>
#include <sys/ioctl.h>
#include <errno.h>
#include "ubsec.h"
#include "ubsio.h"
#include "ubsec_lib.h"


/*
 *  Global Variables:
 */

extern unsigned long user_timing;

/*
 * Function Prototypes: 
 */

int 
eea(int fd,
    unsigned char *n, int *n_bits,        /* the modulus */
    unsigned char *b, int *b_bits,        /* of which we find the inverse */
    unsigned char *t, int *t_bits,
    unsigned char *gcd, int *gcd_bits);   /* Greatest Common Divisor */ 

int
LeftShiftOne(unsigned char *n, int *n_bits);


/*
 * Math acceleration function: This is a generic math function. Command type is
 * passed as a parameter.
 */

int
math_accelerate_ioctl(int ubsec_fd, ubsec_MathCommand_t command,
		      unsigned char *ModN, int *ModN_len, unsigned char *unused, int *unused_len, 
		      unsigned char *ParamA, int *ParamA_len, unsigned char *ParamB, int *ParamB_len,
		      unsigned char *Result, int *Result_len)
{
  ubsec_math_io_t		cmd;
  ubsec_MathCommandParams_pt	params = &cmd.Math;
  int			status = 0;
  
  bzero(&cmd, sizeof(cmd));
  cmd.command = command;
  
  /* Now just generically assign the parameters. No
     type checking required. */
  params->ModN.KeyLength = *ModN_len;
  params->ModN.KeyValue = ModN;
  params->ParamA.KeyLength = *ParamA_len;
  params->ParamA.KeyValue = ParamA;
  params->ParamB.KeyLength = *ParamB_len;
  params->ParamB.KeyValue = ParamB;
  
  /*
   * Result will come back
   */
  params->Result.KeyLength = *Result_len;
  params->Result.KeyValue = Result;
  
  /*
   * Call chip.
   */
  status = ubsec_ioctl (ubsec_fd, (unsigned char *)&cmd, UBSEC_MATH_FUNC);
  
  /*
   * Copy back results.
   */
  if (status == UBSEC_STATUS_SUCCESS) {
    *Result_len = params->Result.KeyLength;
  }
  
  user_timing = cmd.time_us;
  
#ifdef TIMING
  if( user_timing != 0 )
    printf("Speed is %f mbps -- %luus\n", 
	   (8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */
  
  return status;
}


int math_accelerate_dblmodexp_ioctl(int ubsec_fd, ubsec_MathCommand_t command,
	unsigned char *ModN0, int *ModN0_len, unsigned char *ModN1, int *ModN1_len, 
	unsigned char *ParamA0, int *ParamA0_len, unsigned char *ParamA1, int *ParamA1_len, 
	unsigned char *ParamE0, int *ParamE0_len, unsigned char *ParamE1, int *ParamE1_len,
	unsigned char *Result0, int *Result0_len, unsigned char *Result1, int *Result1_len)
{
  ubsec_math_io_t		cmd;
  ubsec_MathCommandParams_pt	params = &cmd.Math;
  int			status = 0;
  
  bzero(&cmd, sizeof(cmd));
  cmd.command = command;
  
  /* Now just generically assign the parameters. No
     type checking required. */
  params->ModN.KeyLength = *ModN0_len;
  params->ModN.KeyValue = ModN0;
  params->ModN2.KeyLength = *ModN1_len;
  params->ModN2.KeyValue = ModN1;
  params->ParamA.KeyLength = *ParamA0_len;
  params->ParamA.KeyValue = ParamA0;
  params->ParamB.KeyLength = *ParamA1_len;
  params->ParamB.KeyValue = ParamA1;
  params->ParamC.KeyLength = *ParamE0_len;
  params->ParamC.KeyValue = ParamE0;
  params->ParamD.KeyLength = *ParamE1_len;
  params->ParamD.KeyValue = ParamE1;
  
  /*
   * Results will come back
   */
  params->Result.KeyLength = *Result0_len;
  params->Result.KeyValue = Result0;
  params->Result2.KeyLength = *Result1_len;
  params->Result2.KeyValue = Result1;
  
  /*
   * Call chip.
   */
  status = ubsec_ioctl (ubsec_fd, (unsigned char *)&cmd, UBSEC_MATH_FUNC);
  
  /*
   * Copy back results.
   */
  if (status == UBSEC_STATUS_SUCCESS) {
    *Result0_len = params->Result.KeyLength;
    *Result1_len = params->Result2.KeyLength;
  }
  
  user_timing = cmd.time_us;
  
#ifdef TIMING
  if( user_timing != 0 )
    printf("Speed is %f mbps -- %luus\n", 
	   (8 * input_size) / (double)user_timing, user_timing);
#endif /* TIMING */
  
  return status;
}


/*
 *  Function: ubsec_modrem
 *
 *  Description: This function computes the remainder of a
 *               number over a modulus, c = a mod n. Unlike
 *               the math_accelerate function or the hardware
 *               this function does not require that |a| = |n|.
 *               That is, a and n can be of different bit-widths.
 *               This function uses the chip's RSA CRT function
 *               to compute c = a mod n.
 *
 *  Return Values:
 *                 == 0  =>  OK
 *                 != 0  =>  Error
 *
 */

int
ubsec_modrem(
	     int fd,
	     unsigned char *a, int *a_bits,
	     unsigned char *n, int *n_bits,
	     unsigned char *c, int *c_bits
	     ) {
  
  int status = 0;
  
  unsigned char *zero = NULL;
  unsigned char *one  = NULL;
  
  int zerobitsdata = *n_bits;
  int onebitsdata  = *n_bits;
  
  int *zero_bits = &zerobitsdata;
  int *one_bits  = &onebitsdata;
  
  do {
    
    if(!(zero = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_modrem: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(zero, MAX_MATH_BYTES);
    
    if(!(one = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_modrem: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(one, MAX_MATH_BYTES);
    one[0] = 0x01;
    
    if((fd = ubsec_open(UBSEC_KEY_DEVICE)) < 0) {
      fprintf(stderr, "error opening device\n");
      status = fd;
      break;
    }
    
    status = rsa_mod_exp_crt_ioctl(fd,
				   a, *a_bits,        /* x */
				   zero, *zero_bits,  /* qinv */
				   one, *one_bits,    /* edq */
				   n, *n_bits,        /* q */
				   one, *one_bits,    /* edp */
				   n, *n_bits,        /* p */
				   c, c_bits);      /* y */

    if(status != 0) {
      return (status);
    }
    
  } while(0);

  if(zero != NULL) {
    bzero(zero, BITSTOBYTES(*zero_bits));
    free(zero);
    zero = NULL;
    *zero_bits = 0;
  }

  if(one != NULL) {
    bzero(one, BITSTOBYTES(*one_bits));
    free(one);
    one = NULL;
    *one_bits = 0;
  }
  
  return(status);
}


/*
 *  Function: ubsec_modinv
 *
 *  Description: This function computes the modulus of a number
 *               over an inverse, t = b ** -1 mod n.
 *
 *  Return Status:
 *                 == 0  =>  OK
 *                 != 0  =>  Error
 *
 */

int
ubsec_modinv(
	     int fd,
	     unsigned char *n, int *n_bits,
	     unsigned char *b, int *b_bits,
	     unsigned char *t, int *t_bits
	     ) {
  
  unsigned char *  gcd = NULL;
  unsigned char *  one = NULL;
 
  int  gcdbitsdata = 0;
  int  onebitsdata = 0;
  
  int *  gcd_bits = &gcdbitsdata;
  int *  one_bits = &onebitsdata;
  
  int  status = 0;

  do {
    
    if(!(gcd = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_modinv: malloc error\n");
      status = -ENOMEM;
      return status;
    }
    bzero(gcd, MAX_MATH_BYTES);
    *gcd_bits = *n_bits;
    
    if(!(one = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_modinv: malloc error\n");
      status = -ENOMEM;
      return status;
    }
    bzero(one, MAX_MATH_BYTES);
    one[0] = 0x01;
    *one_bits = *n_bits;
    
    status = eea(fd,
		 n, n_bits,
		 b, b_bits,
		 t, t_bits,
		 gcd, gcd_bits);
	
    if(status != 0) {
      break;
    }
    
    *one_bits = *gcd_bits;
    if(Compare(one, *one_bits, gcd, *gcd_bits) != 0) {
      fprintf(stderr, "ubsec_modinv: error, inverse doesn't exist.\n");
      status = 1;
      break;
    }

  } while (0);
  
  if(gcd != NULL) {
    bzero(gcd, BITSTOBYTES(*gcd_bits));
    free(gcd);
    gcd = NULL;
    *gcd_bits = 0;
  }
  
  if(one != NULL) {
    bzero(one, BITSTOBYTES(*one_bits));
    free(one);
    one = NULL;
    *one_bits = 0;
  }
  
  return(status);
}


/*
 *  Function: eea
 * 
 *  Description: This function is an implementation of the
 *               Extended Euclidean Algorithm and is being
 *               used to compute the inverse of a number
 *               over a modulus, i.e., t = b ** -1 mod n.
 *
 *  Return Values:
 *                 == 0  =>  OK
 *                 != 0  =>  Error
 *
 */

int
eea(
    int fd,
    unsigned char * n,   int * n_bits,
    unsigned char * b,   int * b_bits,
    unsigned char * t,   int * t_bits,
    unsigned char * gcd, int * gcd_bits
    ) {
  
  unsigned char *u1      = NULL;
  unsigned char *u2      = NULL;
  unsigned char *u3      = NULL;
  unsigned char *t1      = NULL;
  unsigned char *t2      = NULL;
  unsigned char *t3      = NULL;
  unsigned char *zero    = NULL;
  unsigned char *one     = NULL;
  unsigned char *modulus = NULL;
  unsigned char *tmp     = NULL;
  
  int u1bitsdata         = 0;
  int u2bitsdata         = 0;
  int u3bitsdata         = 0;
  int t1bitsdata         = 0;
  int t2bitsdata         = 0;
  int t3bitsdata         = 0;
  int zerobitsdata       = 0;
  int onebitsdata        = 0;
  int modulusbitsdata    = 0;

  int *u1_bits           = &u1bitsdata;
  int *u2_bits           = &u2bitsdata;
  int *u3_bits           = &u3bitsdata;
  int *t1_bits           = &t1bitsdata;
  int *t2_bits           = &t2bitsdata;
  int *t3_bits           = &t3bitsdata;
  int *zero_bits         = &zerobitsdata;
  int *one_bits          = &onebitsdata;
  int *modulus_bits      = &modulusbitsdata;
  
  int status = 0;
  int k      = 0;
  int i      = 0;
  int bits   = 0;
  
  do {
    
    bits = *n_bits + 32;
    *n_bits = bits;
    *b_bits = bits;
    *t_bits = bits;
    
    if(!(zero = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(zero, MAX_MATH_BYTES);
    *zero_bits = bits;

    if(!(one = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(one, MAX_MATH_BYTES);
    one[0] = 0x01;
    *one_bits = bits;

    if(!(modulus = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    memset(modulus, 0xFF, MAX_MATH_BYTES);
    *modulus_bits = bits;

    if(!(u1 = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(u1, MAX_MATH_BYTES);
    *u1_bits = bits;
    
    if(!(u2 = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(u2, MAX_MATH_BYTES);
    *u2_bits = bits;
    
    if(!(u3 = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(u3, MAX_MATH_BYTES);
    *u3_bits = bits;
    
    if(!(t1 = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(t1, MAX_MATH_BYTES);
    *t1_bits = bits;
    
    if(!(t2 = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(t2, MAX_MATH_BYTES);
    *t2_bits = bits;
    
    if(!(t3 = (unsigned char *)malloc(MAX_MATH_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "eaa: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(t3, MAX_MATH_BYTES);
    *t3_bits = bits;

    if(Compare(n, *n_bits, b, *b_bits) <= 0) {
      fprintf(stderr, "error: eea modulus is smaller than number to invert.\n");
      status = 1;
      break;
    }
    
    for(k = 0; ((n[0] & 0x01) == 0) && ((b[0] & 0x01) == 0); k++) {
      fprintf(stderr, "error: error case -- right shifting problem.\n");
      return 1;
      RightShiftOne(n, *n_bits, n, n_bits);
      RightShiftOne(b, *b_bits, b, b_bits);
    }
    
    /* u1 = 1 */
    bzero(u1, BITSTOBYTES(*u1_bits));
    u1[0] = 0x01;

    /* u2 = 0 */
    bzero(u2, BITSTOBYTES(*u2_bits));

    /* u3 = n */
    bzero(u3, BITSTOBYTES(*u3_bits));
    status = math_accelerate_ioctl(fd,                     /* op: u3 = n */
				   UBSEC_MATH_MODADD,
				   modulus, modulus_bits,  /* m */
				   one, one_bits,          /* e */
				   u3, u3_bits,            /* a */
				   n, n_bits,              /* b */
				   u3, u3_bits);           /* c */

    /* t1 = b */
    bzero(t1, BITSTOBYTES(*t1_bits));
    status = math_accelerate_ioctl(fd,                     /* op: t1 = b */
				   UBSEC_MATH_MODADD,
				   modulus, modulus_bits,  /* m */
				   one, one_bits,          /* e */
				   t1, t1_bits,            /* a */
				   b, b_bits,              /* b */
				   t1, t1_bits);           /* c */
    
    /* t2 = n - 1 */
    bzero(t2, BITSTOBYTES(*t2_bits));
    status = math_accelerate_ioctl(fd,                     /* op: t2 = n - 1 */
				   UBSEC_MATH_MODSUB,
				   modulus, modulus_bits,  /* m */
				   one, one_bits,          /* e */
				   n, n_bits,              /* a */
				   one, one_bits,          /* b */
				   t2, t2_bits);           /* c */
    
    /* t3 = b */
    bzero(t3, BITSTOBYTES(*t3_bits));
    status = math_accelerate_ioctl(fd,                     /* op: t3 = b */
				   UBSEC_MATH_MODADD,
				   modulus, modulus_bits,  /* m */
				   one, one_bits,          /* e */
				   t3, t3_bits,            /* a */
				   b, b_bits,              /* b */
				   t3, t3_bits);           /* c */
    
    do {
      do {
	if((u3[0] & 0x01) == 0) {
	  if(((u1[0] & 0x01) == 1) || ((u2[0] & 0x01) == 1)) {
	    status = math_accelerate_ioctl(fd,                     /* op: u1 = u1 + b */
					   UBSEC_MATH_MODADD,
					   modulus, modulus_bits,  /* m */
					   one, one_bits,          /* e */
					   u1, u1_bits,            /* a */
					   b, b_bits,              /* b */
					   u1, u1_bits);           /* c */
	    status = math_accelerate_ioctl(fd,                     /* op: u2 = u2 + n */
					   UBSEC_MATH_MODADD,
					   modulus, modulus_bits,  /* m */
					   one, one_bits,          /* e */
					   u2, u2_bits,            /* a */
					   n, n_bits,              /* b */
					   u2, u2_bits);           /* c */
	  }
	  RightShiftOne(u1, *u1_bits, u1, u1_bits);
	  RightShiftOne(u2, *u2_bits, u2, u2_bits);
	  RightShiftOne(u3, *u3_bits, u3, u3_bits);
	}
	if(((t3[0] & 0x01) == 0) || (Compare(u3, *u3_bits, t3, *t3_bits) < 0)) {
	  tmp = u1, u1 = t1, t1 = tmp;
	  tmp = u2, u2 = t2, t2 = tmp;
	  tmp = u3, u3 = t3, t3 = tmp;
	}
      } while(((u3[0] & 0x01) == 0) && (status == 0));
      
      while(((Compare(u1, *u1_bits, t1, *t1_bits) < 0) || (Compare(u2, *u2_bits, t2, *t2_bits) < 0)) &&
	    (status == 0)) {
	status = math_accelerate_ioctl(fd,                     /* op: u1 = u1 + b */
				       UBSEC_MATH_MODADD,
				       modulus, modulus_bits,  /* m */
				       one, one_bits,          /* e */
				       u1, u1_bits,            /* a */
				       b, b_bits,              /* b */
				       u1, u1_bits);           /* c */
	status = math_accelerate_ioctl(fd,                     /* op: u2 = u2 + n */
				       UBSEC_MATH_MODADD,
				       modulus, modulus_bits,  /* m */
				       one, one_bits,          /* e */
				       u2, u2_bits,            /* a */
				       n, n_bits,              /* b */
				       u2, u2_bits);           /* c */
      }
      
      status = math_accelerate_ioctl(fd,                     /* op: u1 = u1 - t1 */
				     UBSEC_MATH_MODSUB,
				     modulus, modulus_bits,  /* m */
				     one, one_bits,          /* e */
				     u1, u1_bits,            /* a */
				     t1, t1_bits,            /* b */
				     u1, u1_bits);           /* c */
      status = math_accelerate_ioctl(fd,                     /* op: u2 = u2 - t2 */
				     UBSEC_MATH_MODSUB,
				     modulus, modulus_bits,  /* m */
				     one, one_bits,          /* e */
				     u2, u2_bits,            /* a */
				     t2, t2_bits,            /* b */
				     u2, u2_bits);           /* c */
      status = math_accelerate_ioctl(fd,                     /* op: u3 = u3 - t3 */
				     UBSEC_MATH_MODSUB,
				     modulus, modulus_bits,  /* m */
				     one, one_bits,          /* e */
				     u3, u3_bits,            /* a */
				     t3, t3_bits,            /* b */
				     u3, u3_bits);           /* c */
    } while((Compare(t3, *t3_bits, zero, *zero_bits) > 0) && (status == 0));
    
    while(((Compare(u1, *u1_bits, b, *b_bits) >= 0) && (Compare(u2, *u2_bits, n, *n_bits) >= 0))
	  && (status == 0)) {
      
      status = math_accelerate_ioctl(fd,                     /* op: u1 = u1 - b */
				     UBSEC_MATH_MODSUB,
				     modulus, modulus_bits,  /* m */
				     one, one_bits,          /* e */
				     u1, u1_bits,            /* a */
				     b, b_bits,              /* b */
				     u1, u1_bits);           /* c */
      status = math_accelerate_ioctl(fd,                     /* op: u2 = u2 - n */
				     UBSEC_MATH_MODSUB,
				     modulus, modulus_bits,  /* m */
				     one, one_bits,          /* e */
				     u2, u2_bits,            /* a */
				     n, n_bits,              /* b */
				     u2, u2_bits);           /* c */
    }
    
    for(i = 0; i < k; i++) {
      fprintf(stderr, "error: error case -- must fix left shift.\n");
      return 1;
      LeftShiftOne(u1, u1_bits);
      LeftShiftOne(u2, u2_bits);
      LeftShiftOne(u3, u3_bits);
    }

    memcpy(gcd, u3, BITSTOBYTES(*u3_bits));
    status = math_accelerate_ioctl(fd,                     /* op: t = n - u2 */
				   UBSEC_MATH_MODSUB,
				   modulus, modulus_bits,  /* m */
				   one, one_bits,          /* e */
				   n, n_bits,              /* a */
				   u2, u2_bits,            /* b */
				   t, t_bits);             /* c */
    if(status != 0) {
      break;
    }
  } while(0);
  
  if(zero != NULL) {
    bzero(zero, BITSTOBYTES(*zero_bits));
    free(zero);
    zero = NULL;
    *zero_bits = 0;
  }
    
  if(one != NULL) {
    bzero(one, BITSTOBYTES(*one_bits));
    free(one);
    one = NULL;
    *one_bits = 0;
  }
    
  if(modulus != NULL) {
    bzero(modulus, BITSTOBYTES(*modulus_bits));
    free(modulus);
    modulus = NULL;
    *modulus_bits = 0;
  }
    
  if(u1 != NULL) {
    bzero(u1, BITSTOBYTES(*u1_bits));
    free(u1);
    u1 = NULL;
    *u1_bits = 0;
  }
    
  if(u2 != NULL) {
    bzero(u2, BITSTOBYTES(*u2_bits));
    free(u2);
    u2 = NULL;
    *u2_bits = 0;
  }
    
  if(u3 != NULL) {
    bzero(u3, BITSTOBYTES(*u3_bits));
    free(u3);
    u3 = NULL;
    *u3_bits = 0;
  }
    
  if(t1 != NULL) {
    bzero(t1, BITSTOBYTES(*t1_bits));
    free(t1);
    t1 = NULL;
    *t1_bits = 0;
  }
    
  if(t2 != NULL) {
    bzero(t2, BITSTOBYTES(*t2_bits));
    free(t2);
    t2 = NULL;
    *t2_bits = 0;
  }
    
  if(t3 != NULL) {
    bzero(t3, BITSTOBYTES(*t3_bits));
    free(t3);
    t3 = NULL;
    *t3_bits = 0;
  }

  return(status);
}


int
LeftShiftOne(unsigned char *n, int *n_bits)
{
  
  
}

