/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/

/* 
 *  Broadcom Corporation uBSec SDK 
 */

/*
 *  File: ubsec_primerng.c
 *
 *  Desc: Random Prime Number Generator.
 *
 */

/*
 *  Revision History:
 *
 *  12/06/00 DNA Created
 *
 */


#include <stdio.h>
#include <errno.h>
#include "ubsec.h"
#include "ubsio.h"
#include "ubsec_lib.h"


#define NOTPRIME (1)


/*
 * Global Constant Values:
 */

unsigned char   *Zero            = NULL;
unsigned char   *One             = NULL;
unsigned char   *Two             = NULL;
unsigned char   *Modulus         = NULL;

int              ZeroBits        = 0;
int              OneBits         = 0;
int              TwoBits         = 0;
int              ModulusBits     = 0;

int             *ZeroBits_pt     = &ZeroBits;
int             *OneBits_pt      = &OneBits;
int             *TwoBits_pt      = &TwoBits;
int             *ModulusBits_pt  = &ModulusBits;


/*
 * Function Prototypes:
 */

int
InitGlobals(int bits);

int
FreeGlobals(void);

int
TestPrime(int                 fd,
	  ubsec_RNGCommand_t  rng_command,
	  int                 t,
	  unsigned char      *p,
	  int                *p_bits);

int
RightShiftOne(unsigned char  *n, 
	      int             n_bits,
	      unsigned char  *result, 
	      int            *result_bits);

int
Compare(unsigned char *       a, 
	int                   a_bits,
	unsigned char *       b, 
	int                   b_bits);


/*
 * Function: ubsec_primerng
 *
 * Description:
 *              Generates a prime number in result buffer of size 
 *              result_bits.
 *
 */

int
ubsec_primerng(
	       int                 fd,
	       ubsec_RNGCommand_t  rng_command,
	       int                 length,
	       unsigned char      *result,
	       int                *result_bits
	       ) {
  
  int status         = 0;
  int t              = 100;
  int i              = 0;
  
  unsigned char *p           = NULL;
  int prn_bits               = (int)length; /* User-requested number of random bits */
  int *prn_bits_pt           = &prn_bits;
  int ms_uint32              = (ROUNDUP_TO_32_BIT(length)/8) - 1; /* Index to last LongKey element */ 
  
  do {
    
    if((status = InitGlobals(length)) != 0) {
      prn_bits = 0;
      break;
    }

    /* Allocate memory for maximum sized RN */
    if(!(p = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "prime_rng_ioctl: malloc error\n");
      status = -ENOMEM;
      prn_bits = 0;
      break;
    }
    bzero(p, MAX_RNG_BYTES);

    /* The RNG ioctl returns a RN no bigger than that requested.     */
    /* However, the returned size may be smaller than that requested */
    /* if the significant bit(s) = 0.                                */ 
    if((status = rng_ioctl(fd, rng_command, p, prn_bits_pt)) != 0) {
      fprintf(stderr, "prime_rng_ioctl: error in call to rng_ioctl\n");
      prn_bits = 0;
      break;
    }

    /* Guarantee RN is exact requested width and is odd */
    ((UBS_UINT32 *)p)[ms_uint32] |= (0x00000001 << ((length-1) & 31)); 
    ((UBS_UINT32 *)p)[0] |= 0x00000001;
    prn_bits = length;
    
    while(status == 0) {
      if(TestPrime(fd, rng_command, t, p, prn_bits_pt) == 0) {
	/* Got prime RN, break out of loop */
	break;
      } 
      else {
	/* RN not prime */
	if((++i == 1000)) {
	  /* Adjust RN by adding 2 to it */
	  math_accelerate_ioctl(fd,                     /* op: result = result + 2 */
				UBSEC_MATH_MODADD,
				Modulus, ModulusBits_pt,  /* m */
				One, OneBits_pt,          /* e */
				p, prn_bits_pt,           /* a */
				Two, TwoBits_pt,          /* b */ 
				p, prn_bits_pt);          /* c */
	} /* RN adjust */
	else {
	  /* 999 attempts to adjust RN failed; just get another RN */
	  if((status = rng_ioctl(fd, rng_command, p, prn_bits_pt)) != 0) {
	    fprintf(stderr, "prime_rng_ioctl: error in call to rng_ioctl\n");
	    prn_bits = 0;
	    break;
	  }
	  /* Guarantee RN is exact requested width and is odd */
	  ((UBS_UINT32 *)p)[ms_uint32] |= (0x00000001 << ((length-1) & 31)); 
	  ((UBS_UINT32 *)p)[0] |= 0x00000001;
	  prn_bits = length;
	  i = 0; /* Work on this new RN another 999 times (max) */
	} /* Got a brand new RN */
      } /* RN adjust / new RN */
    } /* Until ioctl failure or until prime RN found */
  } while(0);
  
  memcpy(result, p, ROUNDUP_TO_32_BIT(prn_bits)/8);
  
  *result_bits = (unsigned int)prn_bits;

  FreeGlobals();
  
  if(p != NULL) {
    bzero(p, MAX_RNG_BYTES);
    free(p);
    p = NULL;
    prn_bits = 0;
  }

  return(status);
}


/*
 * Function: TestPrime
 *
 * Description:
 *
 * Return Value:
 *
 *  == 0  =>  p is prime
 *  != 0  =>  p is not prime
 */

int
TestPrime(
	  int                 fd,
	  ubsec_RNGCommand_t  rng_command,
	  int                 t,
	  unsigned char *     p, 
	  int                *p_bits
	  ) {
  
  int i, ii;
  int status = 0;

  unsigned char *  a               = NULL;
  unsigned char *  b               = NULL;
  unsigned char *  j               = NULL;
  unsigned char *  m               = NULL;
  unsigned char *  p1              = NULL;
  unsigned char *  z               = NULL;
  
  int              abitsdata       = 0;
  int              bbitsdata       = 0;
  int              jbitsdata       = 0;
  int              mbitsdata       = 0;
  int              p1bitsdata      = 0;
  int              zbitsdata       = 0;
  
  int *            a_bits          = &abitsdata;
  int *            b_bits          = &bbitsdata;
  int *            j_bits          = &jbitsdata;
  int *            m_bits          = &mbitsdata;
  int *            p1_bits         = &p1bitsdata;
  int *            z_bits          = &zbitsdata;

  do {
#if 0 /* Done in InitGlobals() */
    if(!(Modulus = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "TestPrime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    memset(Modulus, 0xFF, MAX_RNG_BYTES);
    ModulusBits = *p_bits;
#endif

    if(!(a = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "testprime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(a, MAX_RNG_BYTES);
    *a_bits = *p_bits;

    if(!(b = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "testprime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(b, MAX_RNG_BYTES);
    *b_bits = *p_bits;
    
    if(!(j = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "testprime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(j, MAX_RNG_BYTES);
    *j_bits = *p_bits;

    if(!(m = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "testprime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(m, MAX_RNG_BYTES);
    *m_bits = *p_bits;
	
    if(!(p1 = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "testprime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(p1, MAX_RNG_BYTES);
    *p1_bits = *p_bits;

    if(!(z = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "testprime: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(z, MAX_RNG_BYTES);
    *z_bits = *p_bits;
	
    /* Compute b such that 2 ** b is the largest power of 2 that divides p - 1 */
    /* Compute m such that p = 1 + 2 ** b * m */
    
    status = math_accelerate_ioctl(fd,                     /* op: m = p - 1 */
				   UBSEC_MATH_MODSUB,
				   Modulus, ModulusBits_pt,  /* m */
				   One, OneBits_pt,          /* e */
				   p, p_bits,                /* a */
				   One, OneBits_pt,          /* b */
				   m, m_bits);               /* c */
    if(status != 0)
      break;
    
    do {
      RightShiftOne(m, *m_bits, m, m_bits);
      status = math_accelerate_ioctl(fd,                     /* op: b = b + 1 */
				     UBSEC_MATH_MODADD, 
				     Modulus, ModulusBits_pt,  /* m */
				     One, OneBits_pt,          /* e */
				     b, b_bits,                /* a */
				     One, OneBits_pt,          /* b */
				     b, b_bits);               /* c */
    } while(((m[0] & 0x01) == 0) && (status == 0));
    
    if(status != 0)
      break;

    status = math_accelerate_ioctl(fd,                     /* op: p1 = p - 1 */
				   UBSEC_MATH_MODSUB,
				   Modulus, ModulusBits_pt,  /* m */
				   One, OneBits_pt,          /* e */
				   p, p_bits,                /* a */
				   One, OneBits_pt,          /* b */
				   p1, p1_bits);             /* c */

    if(status != 0)
      break;
    
    for(i = 0; (i < t) && (status == 0); i++) {
      
      /* !!! must have a < p; for now hack it with bits-1 */
      
      *a_bits -= 1;
      if((status = rng_ioctl(fd, rng_command, a, a_bits)) != 0)
	break;
      *a_bits = *p_bits;

      bzero(j, BITSTOBYTES(*j_bits));
    
      status = math_accelerate_ioctl(fd,                 /* op: z = a ** m mod p */
				     UBSEC_MATH_MODEXP,
				     p, p_bits,          /* m */
				     m, m_bits,          /* e */
				     a, a_bits,          /* a */
				     m, m_bits,          /* b */
				     z, z_bits);         /* c */
      if(status != 0)
	break;
      
      for(ii = 0; (ii < 100000) && (status == 0); ii++) {
	
	if(Compare(z, *z_bits, One, OneBits) == 0) {
	  break;  /* p passes and may be prime */
	}
	
	if(Compare(z, *z_bits, p1, *p1_bits) == 0) {
	  break;  /* p passes and may be prime */
	}

	if((Compare(j, *j_bits, Zero, ZeroBits) > 0) && (Compare(z, *z_bits, One, OneBits) == 0)) {
	  status = NOTPRIME;  /* p fails the test and is not prime */
	  break;
	}

	status = math_accelerate_ioctl(fd,                     /* op: j = j + 1 */
				       UBSEC_MATH_MODADD,
				       Modulus, ModulusBits_pt,  /* m */
				       One, OneBits_pt,          /* e */
				       j, j_bits,                /* a */
				       One, OneBits_pt,          /* b */
				       j, j_bits);               /* c */
	
	if(status != 0)
	  break;
	
	if((Compare(j, *j_bits, b, *b_bits) < 0) && (Compare(z, *z_bits, p1, *p1_bits) != 0)) {
	  status = math_accelerate_ioctl(fd,                   /* op: z = z ** 2 mod p */
					 UBSEC_MATH_MODEXP,
					 p, p_bits,            /* m */
					 Two, TwoBits_pt,      /* e */ 
					 z, z_bits,            /* a */
					 One, OneBits_pt,      /* b */
					 z, z_bits);           /* c */
	  if(status != 0)
	    break;
	  continue;
	}
	
	if((Compare(j, *j_bits, b, *b_bits) == 0) && (Compare(z, *z_bits, p1, *p1_bits) != 0)) {
	  status = NOTPRIME;  /* p fails the test and is not prime */
	  break;
	}
      }
    }
  } while(0);
  
  if(a != NULL) {
    bzero(a, BITSTOBYTES(*a_bits));
    free(a);
    a = NULL;
    *a_bits = 0;
  }
  
  if(b != NULL) {
    bzero(b, BITSTOBYTES(*b_bits));
    free(b);
    b = NULL;
    *b_bits = 0;
  }
  
  if(j != NULL) {
    bzero(j, BITSTOBYTES(*j_bits));
    free(j);
    j = NULL;
    *j_bits = 0;
  }

  if(m != NULL) {
    bzero(m, BITSTOBYTES(*m_bits));
    free(m);
    m = NULL;
    *m_bits = 0;
  }
  
  if(p1 != NULL) {
    bzero(p1, BITSTOBYTES(*p1_bits));
    free(p1);
    p1 = NULL;
    *p1_bits = 0;
  }

  if(z != NULL) {
    bzero(z, BITSTOBYTES(*z_bits));
    free(z);
    z = NULL;
    *z_bits = 0;
  }

  return(status);
}


int
InitGlobals(int bits) {
  
  int status = 0;
  
  do {
    if(!(Zero = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "InitGlobals: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(Zero, MAX_RNG_BYTES);
    ZeroBits = bits;
    
    if(!(One = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "InitGlobals: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(One, MAX_RNG_BYTES);
    ((UBS_UINT32 *)One)[0] = 0x00000001;
    OneBits = bits;
    
    if(!(Two = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "InitGlobals: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(Two, MAX_RNG_BYTES);
    ((UBS_UINT32 *)Two)[0] = 0x00000002;
    TwoBits = bits;
    
    if(!(Modulus = (unsigned char *)malloc(MAX_RNG_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "InitGlobals: malloc error\n");
      status = -ENOMEM;
      break;
    }
    memset(Modulus, 0xFF, MAX_RNG_BYTES);
    ModulusBits = bits;
    
  } while(0);
  
  return(status);
}


int
FreeGlobals(void) {
  
  int status = 0;
  
  if(Zero != NULL) {
    bzero(Zero, BITSTOBYTES(ZeroBits));
    free(Zero);
    Zero = NULL;
    ZeroBits = 0;
  }

  if(One != NULL) {
    bzero(One, BITSTOBYTES(OneBits));
    free(One);
    One = NULL;
    OneBits = 0;
  }

  if(Two != NULL) {
    bzero(Two, BITSTOBYTES(TwoBits));
    free(Two);
    Two = NULL;
    TwoBits = 0;
  }

  if(Modulus != NULL) {
    bzero(Modulus, BITSTOBYTES(ModulusBits));
    free(Modulus);
    Modulus = NULL;
    ModulusBits = 0;
  }

  return(status);
}


int
RightShiftOne(
	      unsigned char *n, int n_bits,
	      unsigned char *result, int *result_bits
	      ) 
{
  int element = ((ROUNDUP_TO_32_BIT(n_bits)) / 32) - 1;
  UBS_UINT32 shifted_bit = 0;
  
  if (n_bits <= 0)
    return (*result_bits = 0);

  for( ; element >= 0; element--) {
    ((UBS_UINT32 *)result)[element] = (((UBS_UINT32 *)n)[element] >> 1) | shifted_bit;
    shifted_bit = (((UBS_UINT32 *)n)[element] & 0x00000001) << 31;
  }
  *result_bits = n_bits - 1;
}


/*
 * Return values:
 *
 * a < b   =>  Return value is negative
 * a == b  =>  Return value is zero
 * a > b   =>  Return value is positive
 *
 */

int
Compare(
	unsigned char *a, int a_bits,
	unsigned char *b, int b_bits
	) {
  
  int i;
  
  if(BITSTOBYTES(a_bits) != BITSTOBYTES(b_bits)) {
    /* fprintf(stderr, "Compare bits not equal: %d %d\n", a_bits, b_bits); */
    return (a_bits - b_bits);
  }
  
  for(i = BITSTOBYTES(a_bits) - 1; i >= 0; i--) {
    if(a[i] < b[i]) 
      return -1;
    if(a[i] > b[i])
      return 1;
  }
  return 0;
}


int
PrintNumber(FILE *ofptr, unsigned char *num, unsigned int bits) 
{
  int element = ((ROUNDUP_TO_32_BIT(bits)) / 32) -1 ;
  int i = 0;
  
  for( ; element >= 0; element--) {
    if (((i%8) == 7) || !element)
      fprintf(ofptr, "%08X\n", ((UBS_UINT32 *)num)[element]);
    else
      fprintf(ofptr, "%08X ", ((UBS_UINT32 *)num)[element]);
  }
  fprintf(ofptr, "\n");
}











