/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/

/* 
 *  Broadcom Corporation uBSec SDK 
 */

/*
 *  File: ubsec_rsakeygen.c
 *
 *  Desc: RSA Key Generation Functions.
 *
 */

/*
 *  Revision History:
 *
 *  12/06/00 DNA Created.
 *
 */


#include <stdio.h>
#include <errno.h>
#include "ubsec.h"
#include "ubsio.h"
#include "ubsec_lib.h"


/*
 *  Function: ubsec_rsakeygen
 *
 *  Description: This function takes as its inputs a public
 *               exponent and the lengths of the private
 *               values p and q. From these, this function
 *               computes all of the public and private
 *               keys necessary for RSA Public and Private
 *               CRT operations.
 *
 *  Return Values:
 *                 == 0  =>  OK
 *                 != 0  =>  Error
 */

int
ubsec_rsakeygen(
		int fd,
		int modulus_bits,
		unsigned char *  e,    int *  e_bits,
		unsigned char *  p,    int *  p_bits,
		unsigned char *  q,    int *  q_bits,
		unsigned char *  n,    int *  n_bits,
		unsigned char *  d,    int *  d_bits,
		unsigned char *  dp,   int *  dp_bits,
		unsigned char *  dq,   int *  dq_bits,
		unsigned char *  pinv, int *  pinv_bits
		) {
  
  int status = 0;
  
  unsigned char *  p1       = NULL;
  unsigned char *  q1       = NULL;
  unsigned char *  p1q1     = NULL;
  unsigned char *  p1q12    = NULL;
  unsigned char *  zero     = NULL;
  unsigned char *  one      = NULL;
  unsigned char *  two      = NULL;
  unsigned char *  nmodulus = NULL;
  unsigned char *  pmodulus = NULL;
  unsigned char *  tmp      = NULL;
  
  int  p1bitsdata       = 0;
  int  q1bitsdata       = 0;
  int  p1q1bitsdata     = 0;
  int  p1q12bitsdata    = 0;
  int  zerobitsdata     = 0;
  int  onebitsdata      = 0;
  int  twobitsdata      = 0;
  int  nmodulusbitsdata = 0;
  int  pmodulusbitsdata = 0;
  int  tmpbitsdata      = 0;
  
  int    *p1_bits       = &p1bitsdata;
  int    *q1_bits       = &q1bitsdata;
  int    *p1q1_bits     = &p1q1bitsdata;
  int    *p1q12_bits    = &p1q12bitsdata;
  int    *zero_bits     = &zerobitsdata;
  int    *one_bits      = &onebitsdata;
  int    *two_bits      = &twobitsdata;
  int    *nmodulus_bits = &nmodulusbitsdata;
  int    *pmodulus_bits = &pmodulusbitsdata;
  int    *tmp_bits      = &tmpbitsdata;
  

  do {

    /*
     * Restriction: modulus_bits must be double word aligned as yet...
     */

    if(modulus_bits % 64 != 0) {
      fprintf(stderr, "ubsec_rsakeygen error: modulus_bits must be double word aligned\n");
      status = 1;
      break;
    }

    *p_bits = modulus_bits / 2;
    *q_bits = modulus_bits / 2;


    if(!(p1 = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(p1, MAX_PUBLIC_KEY_BYTES);
    *p1_bits = *p_bits;
    
    if(!(q1 = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(q1, MAX_PUBLIC_KEY_BYTES);
    *q1_bits = *q_bits;
    
    if(!(p1q1 = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(p1q1, MAX_PUBLIC_KEY_BYTES);
    *p1q1_bits = *n_bits;
    
    if(!(p1q12 = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(p1q12, MAX_PUBLIC_KEY_BYTES);
    *p1q12_bits = *n_bits;
    
    if(!(zero = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(zero, MAX_PUBLIC_KEY_BYTES);
    *zero_bits = 32;

    if(!(one = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(one, MAX_PUBLIC_KEY_BYTES);
    one[0] = 0x01;
    *one_bits = 32;

    if(!(two = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(two, MAX_PUBLIC_KEY_BYTES);
    two[0] = 0x02;
    *two_bits = 32;

    if(!(nmodulus = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    memset(nmodulus, 0xFF, MAX_PUBLIC_KEY_BYTES);
    *nmodulus_bits = *n_bits;
    
    if(!(pmodulus = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    memset(pmodulus, 0xFF, MAX_PUBLIC_KEY_BYTES);
    *pmodulus_bits = *p_bits;
    
    if(!(tmp = (unsigned char *)malloc(MAX_PUBLIC_KEY_BYTES * sizeof(unsigned char)))) {
      fprintf(stderr, "ubsec_rsakeygen: malloc error\n");
      status = -ENOMEM;
      break;
    }
    bzero(tmp, MAX_PUBLIC_KEY_BYTES);
    *tmp_bits = *n_bits;
    
    /*****************
     * Compute p and q
     *****************/
    
    status = ubsec_primerng(fd,
			    UBSEC_RNG_SHA1,
			    *p_bits,
			    p,
			    p_bits);
    
    status = ubsec_primerng(fd,
			    UBSEC_RNG_SHA1,
			    *q_bits,
			    q,
			    q_bits);

    /*******************
     * Compute n = p * q
     *******************/
    
    status = math_accelerate_ioctl(fd,                       /* op: n = p * q */
				   UBSEC_MATH_MODMUL,
				   nmodulus, nmodulus_bits,  /* m */
				   p, p_bits,                /* e */
				   p, p_bits,                /* a */
				   q, q_bits,                /* b */
				   n, n_bits);               /* c */
    if(status != 0) {
      break;
    }

    /******************************************
     * Compute d = e ** -1 mod ((p - 1)(q - 1))
     ******************************************/
    
    status = math_accelerate_ioctl(fd,                       /* op: p1 = p - 1 */
				   UBSEC_MATH_MODSUB,
				   pmodulus, pmodulus_bits,  /* m */
				   p, p_bits,                /* e */
				   p, p_bits,                /* a */
				   one, one_bits,            /* b */
				   p1, p1_bits);             /* c */
    if(status != 0) {
      break;
    }

    status = math_accelerate_ioctl(fd,                       /* op: q1 = q - 1 */
				   UBSEC_MATH_MODSUB,
				   pmodulus, pmodulus_bits,  /* m */
				   q, q_bits,                /* e */
				   q, q_bits,                /* a */
				   one, one_bits,            /* b */
				   q1, q1_bits);             /* c */
    if(status != 0) {
      break;
    }
    
    status = math_accelerate_ioctl(fd,                       /* op: p1q1 = p1 * q1 */
				   UBSEC_MATH_MODMUL,
				   nmodulus, nmodulus_bits,  /* m */
				   p, p_bits,                /* e */
				   p1, p1_bits,              /* a */
				   q1, q1_bits,              /* b */
				   p1q1, p1q1_bits);         /* c */
    if(status != 0) {
      break;
    }

    status = ubsec_modinv(fd,
			  p1q1, p1q1_bits,
			  e, e_bits,
			  d, d_bits);
    if(status != 0) {
      break;
    }

    /****************************
     * Compute dp = d mod (p - 1)
     ****************************/
    
    status = ubsec_modrem(fd,
			  d, d_bits,
			  p1, p1_bits,
			  dp, dp_bits);
    if(status != 0) {
      break;
    }
    
    /****************************
     * Compute dq = d mod (q - 1)
     ***************************/
    
    status = ubsec_modrem(fd,
			  d, d_bits,
			  q1, q1_bits,
			  dq, dq_bits);
    if(status != 0) {
      break;
    }
    
    /******************************
     * Compute pinv = p ** -1 mod q
     ******************************/
    
    bzero(tmp, MAX_PUBLIC_KEY_BYTES);
    status = ubsec_modrem(fd,
			  p, p_bits,
			  q, q_bits,
			  tmp, tmp_bits);
    if(status != 0) {
      break;
    }
    
    status = ubsec_modinv(fd,
			  q, q_bits,
			  tmp, tmp_bits,
			  pinv, pinv_bits);
    if(status != 0) {
      break;
    }
    
  } while(0);
  
  if(p1 != NULL) {
    bzero(p1, BITSTOBYTES(*p1_bits));
    free(p1);
    p1 = NULL;
    *p1_bits = 0;
  }

  if(q1 != NULL) {
    bzero(q1, BITSTOBYTES(*q1_bits));
    free(q1);
    q1 = NULL;
    *q1_bits = 0;
  }

  if(p1q1 != NULL) {
    bzero(p1q1, BITSTOBYTES(*p1q1_bits));
    free(p1q1);
    p1q1 = NULL;
    *p1q1_bits = 0;
  }

  if(p1q12 != NULL) {
    bzero(p1q12, BITSTOBYTES(*p1q12_bits));
    free(p1q12);
    p1q12 = NULL;
    *p1q12_bits = 0;
  }

  if(zero != NULL) {
    bzero(zero, BITSTOBYTES(*zero_bits));
    free(zero);
    zero = NULL;
    *zero_bits = 0;
  }

  if(one != NULL) {
    bzero(one, BITSTOBYTES(*one_bits));
    free(one);
    one = NULL;
    *one_bits = 0;
  }
 
  if(two != NULL) {
    bzero(two, BITSTOBYTES(*two_bits));
    free(two);
    two = NULL;
    *two_bits = 0;
  }

  if(nmodulus != NULL) {
    bzero(nmodulus, BITSTOBYTES(*nmodulus_bits));
    free(nmodulus);
    nmodulus = NULL;
    *nmodulus_bits = 0;
  }

  if(pmodulus != NULL) {
    bzero(pmodulus, BITSTOBYTES(*pmodulus_bits));
    free(pmodulus);
    pmodulus = NULL;
    *pmodulus_bits = 0;
  }

  if(tmp != NULL) {
    bzero(tmp, BITSTOBYTES(*tmp_bits));
    free(tmp);
    tmp = NULL;
    *tmp_bits = 0;
  }
  
  return(status);
}


