/*
 *  Broadcom Cryptonet Driver software is distributed as is, without any warranty
 *  of any kind, either express or implied as further specified in the GNU Public
 *  License. This software may be used and distributed according to the terms of
 *  the GNU Public License.
 *
 * Cryptonet is a registered trademark of Broadcom Corporation.
 */
/******************************************************************************
 *
 *  Copyright 2000
 *  Broadcom Corporation
 *  16215 Alton Parkway
 *  PO Box 57013
 *  Irvine CA 92619-7013
 *
 *****************************************************************************/     

/*******************************************************************************
 *
 * File:         ubslib/ubecom_ssl_ioctl.c
 * 
 * What:         SSL/TLS additions to the Crypto Application Library
 *
 * Description:
 *              The following are a set of function calls corresponding to the
 *              SSL and TLS functionality of the BCM582x. Generally, these 
 *              functions receive input data from the user, package this data up 
 *              into structures defined in Linux/ubsio.h and make an ioctl call 
 *              to the SRL and the chip.
 * 
 * Revision History:
 *                   When       Who   What
 *                   09/05/00   DNA   Created
 *
 ******************************************************************************/

#include <stdio.h>
#include <sys/ioctl.h>
#include "ubsec.h"
#include "ubsio.h"
#include "ubsec_lib.h"

/*
 * Global Variables:
 */

extern unsigned long user_timing;


/**************************************************************************
 *
 * Function:    ubecom_ssl_mac_ioctl
 * 
 * Called From: Any user application.
 *
 * What:        Interface between user and OS for an SSL MAC operation.
 * 
 * Description: 
 *              Receives from the user all input data pertinant to an SSL 
 *              MAC operation, inserts this data into a ubecom_ssl_io_t
 *              struct and makes the ioctl call through ubsec_ioctl.
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_ssl_mac_ioctl(
		    int              fd,
		    unsigned char *  data_in,      /* data to authenticate */
		    unsigned int     data_in_len,  /* length of data to auth */
		    unsigned char *  auth_key,     /* secret key */
		    unsigned int     auth_key_len, /* length of secret key */
		    unsigned char *  seq_num,      /* sequence number */
		    unsigned char    type,         /* content type */
		    unsigned short   data_length,  /* data length */
		    unsigned int     auth_alg,     /* hash agthm: sha or md5 */
		    unsigned char *  data_out,     /* buffer for output data */
		    unsigned int     data_out_len  /* length of output buffer */
		    ) {
  
  int                status  = -1;
  ubsec_sslmac_io_t  cmd;
  
  bzero(&cmd, sizeof(ubsec_sslmac_io_t));

  cmd.command           = UBSEC_SSL_MAC;
  cmd.SourceBuffer      = data_in;
  cmd.SourceBufferBytes = data_in_len;
  cmd.HmacBuffer        = data_out;
  cmd.HmacBufferBytes   = data_out_len;
  cmd.HashAlgorithm     = auth_alg;
  cmd.ContentType       = type;
  cmd.DataLength        = data_length;
  memcpy(&(cmd.SequenceNumber), seq_num, sizeof(ubsec_DoubleSequenceNumber_t));
  memcpy(&(cmd.key), auth_key, sizeof(ubsec_SSLMAC_key_t));
  
  status = ubsec_ioctl(fd, (unsigned char *)&cmd, UBSEC_SSL_MAC_FUNC);
  
  user_timing = cmd.time_us;

  return(status);
}


/**************************************************************************
 *
 * Function:    ubecom_tls_mac_ioctl
 * 
 * Called From: Any user application.
 *
 * What:        Interface between user and OS for an SSL MAC operation.
 * 
 * Description: 
 *              Receives from the user all input data pertinant to a TLS 
 *              MAC operation, inserts this data into a ubecom_ssl_io_t
 *              struct and makes the ioctl call through ubsec_ioctl.
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_tls_mac_ioctl(
		    int              fd,
		    unsigned char *  data_in,      /* data to authenticate */
		    unsigned int     data_in_len,  /* length of data to auth */
		    unsigned char *  auth_key,     /* MAC Secret Key */
		    unsigned int     auth_key_len, /* MAC Secret Key in Bytes */
		    unsigned char *  seq_num,      /* sequence number */
		    unsigned long    version,      /* TLS Version */
		    unsigned char    type,         /* TLS Content Type */
		    unsigned long    datalength,   /* data length */
		    unsigned int     auth_alg,     /* hash agthm: sha or md5 */
		    unsigned char *  data_out,     /* buffer for output data */
		    unsigned int     data_out_len  /* length of output buffer */
		    ) {
  
  int                status = 0;
  ubsec_tlsmac_io_t  cmd;
 
  bzero(&cmd, sizeof(cmd));
 
  cmd.command           = UBSEC_TLS | auth_alg;
  cmd.SourceBuffer      = data_in;
  cmd.SourceBufferBytes = data_in_len;
  cmd.HmacBuffer        = data_out;
  cmd.HmacBufferBytes   = data_out_len;
  cmd.HashAlgorithm     = auth_alg;
  cmd.ContentType       = type;
  cmd.DataLength        = (unsigned short)(0xFFFF & datalength);
  cmd.Version           = (unsigned short)(0xFFFF & version);
 
  memcpy(&(cmd.SequenceNumber), seq_num, sizeof(ubsec_DoubleSequenceNumber_t));
  memcpy(&(cmd.MacKey), auth_key, auth_key_len);
  
  status = ubsec_ioctl(fd, (unsigned char *)&cmd, UBSEC_TLS_HMAC_FUNC);
  
  user_timing = cmd.time_us;

  return(status);
}


/**************************************************************************
 *
 * Function:    ubecom_ssl_cipher_ioctl
 * 
 * Called From: Any user application.
 *
 * What:        Interface between user and OS for an SSL 3DES operation.
 * 
 * Description:
 *              Receives from the user all input data pertinant to an SSL 
 *              3DES operation, inserts this data into a ubecom_ssl_io_t
 *              struct and makes the ioctl call through ubsec_ioctl.
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_ssl_cipher_ioctl(
		       int fd,
		       unsigned char *  data_in,      /* input data buffer */
		       unsigned int     data_in_len,  /* length input buffer */
		       unsigned char *  key1,         /* first DES key */
		       unsigned char *  key2,         /* second DES key */
		       unsigned char *  key3,         /* third DES key */
		       unsigned char *  iv,           /* initial vector */
		       unsigned int     encodeflag,   /* encrypt or decrypt */
		       unsigned char *  data_out,     /* output data buffer */
		       unsigned int     data_out_len  /* length of output buffer */
		       ) {
  
  int                   status = -1;
  ubsec_sslcipher_io_t  cmd;
  
  bzero(&cmd, sizeof(cmd));

  cmd.command = (encodeflag) ? UBSEC_SSL_DES_ENCODE : UBSEC_SSL_DES_DECODE;
  
  memcpy(&(cmd.CryptKey[0]),  key1, 8);
  memcpy(&(cmd.CryptKey[8]),  key2, 8);
  memcpy(&(cmd.CryptKey[16]), key3, 8);
  memcpy(&(cmd.InitialVector), iv, IV_LENGTH);

  cmd.SourceBuffer      = data_in;
  cmd.SourceBufferBytes = data_in_len;
  cmd.DestBuffer        = data_out;
  cmd.DestBufferBytes   = data_out_len;
  
  status = ubsec_ioctl(fd, (unsigned char *)&cmd, UBSEC_SSL_DES_FUNC);
  
  user_timing = cmd.time_us;
  
  return(status);
}


/**************************************************************************
 *
 * Function:    ubsec_ssl_arc4_init_ioctl
 * 
 * Called From: Any user application.
 *
 * What:        Initialization function to be called before a call to
 *              ubsec_ssl_arc4_process_ioctl when starting from key data.
 * 
 * Description: 
 *              Receives key and index data for an SSL ARC4 operation, 
 *              inserts this data into the key_buf buffer in a format that
 *              prepares the data for a call to 
 *              ubsec_ssl_arc4_process_ioctl. 
 *              The input key should be anywhere from 1 to 256 bytes 
 *              long. This function is responsible for repeating the 
 *              key up to 256 bytes. 
 *              The key_buf array should be 260 bytes long.
 *              The format of key_buf is this:
 *              byte[0] = index_j, byte[2] = index_i, byte[1] and byte[3]
 *              are reserved; byte[4] through byte[255] hold key data.
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_ssl_arc4_init_ioctl(
			  int fd,
			  unsigned char *  key,        /* key data */
			  unsigned int     key_len,    /* length in bytes of key */
			  unsigned char    index_i,    /* initial value of index i */
			  unsigned char    index_j,    /* initial value of index j */
			  unsigned char *  key_buf,    /* output key/index buffer */
			  unsigned int     key_buf_len /* length of output key/index buffer in bytes */
			  ) {
  int i;
  
  if((key_len < 1) || (key_len > 256)) {
    fprintf(stderr, "ubsec_ssl_arc4_init_ioctl: error key_len = %d\n", key_len);
    return(-1);
  }

  if(key_buf_len != 260) {
    fprintf(stderr, "ubsec_ssl_arc4_init_ioctl: error key_buf_len = %d\n", key_buf_len);
    return(-1);
  }
  
  bzero(key_buf, key_buf_len);

  key_buf[0] = index_j;
  key_buf[2] = index_i;
  
  /* repeat a key of length less than 256 bytes */
  
  for(i = 0; i < 256; i++) {
    (key_buf + 4)[i] = key[i % key_len];
  }
  
  return(0);
}


/**************************************************************************
 *
 * Function:    ubsec_ssl_arc4_process_ioctl
 * 
 * Called From: Any user application.
 *
 * What:        Interface between user and OS for an SSL ARC4 operation.
 * 
 * Description: 
 *              Receives from the user all input data pertinant to an SSL 
 *              ARC4 operation, inserts this data into a ubsec_arc4_io_t
 *              struct and makes the ioctl call through ubsec_ioctl.
 *              The key_state buffer should hold either the output of a
 *              call to ubsec_ssl_arc4_init_ioctl or the output of a
 *              previous call to ubsec_ssl_arc4_process_ioctl.
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_ssl_arc4_process_ioctl(
			     int fd,
			     unsigned char *data_in,      /* input data buffer */
			     unsigned int data_in_len,    /* length of input buffer */
			     unsigned char *key_state,    /* key or state */
			     unsigned int key_state_flag, /* start from key or state */
			     unsigned char *data_out,     /* output data buffer */
			     unsigned int data_out_len,   /* length of output buffer */
			     unsigned char *state_out     /* state upon completion */
			     ) {
  
  ubsec_arc4_io_t  cmd;
  int              status = 0;
  
  bzero(&cmd, sizeof(cmd));
  
  cmd.command               = key_state_flag;
  cmd.num_fragments         = 1;
  cmd.SourceBuffer          = data_in;
  cmd.SourceBufferBytes     = data_in_len;
  cmd.KeyStateFlag          = key_state_flag;
  cmd.KeyStateInBuffer      = key_state;
  cmd.KeyStateInBufferBytes = ARC4_INDICESKEYSTATE_BUF_LEN;
  cmd.DestBuffer            = data_out;
  cmd.DestBufferBytes       = data_out_len;
  cmd.StateOutBuffer        = state_out;
  cmd.StateOutBufferBytes   = ARC4_INDICESKEYSTATE_BUF_LEN;

  status = ubsec_ioctl(fd, (unsigned char *)&cmd, UBSEC_SSL_ARC4_FUNC);


#ifdef REQ_XOR_IN__NULLDATAMODE
	/* Doing XOR for Null Data Mode for 5821 */
	{
	#define XOR_TYPE long

  	XOR_TYPE *srcptr=NULL,*destptr=NULL;
	int  num=0, i=0;
	if ( (key_state_flag & ARC4_NULL_DATA_MODE) == ARC4_NULL_DATA_MODE)
		{
		num = data_out_len/sizeof(XOR_TYPE);
		srcptr = (long *) data_in;
		destptr = (long *) data_out;
        	for(i=0;i<=num;i++)
			{
			*destptr++ ^= *srcptr++;
			}
		num *= sizeof(XOR_TYPE);
        	for(i=num;i<=data_out_len;i++)
			 data_out[i] ^= data_in[i];
		}
	}
#endif
  
  user_timing = cmd.time_us;
  
  return(status);
}


/**************************************************************************
 *
 * Function:    ubsec_hash_ioctl
 * 
 * Called From: Any user application.
 *
 * What:        Interface between user and OS for a Hash operation.
 * 
 * Description: 
 *
 * Return Values: 
 *                == 0   =>   Success
 *                != 0   =>   Failure
 *
 *************************************************************************/

int
ubsec_hash_ioctl(
		 int fd,
		 unsigned char *data_in,          /* input data buffer */
		 unsigned int data_in_len,        /* length of input data buffer */
		 int auth_alg,                    /* sha or md5 hash alg */
		 unsigned char *data_out,         /* output data buffer */
		 unsigned int data_out_len        /* length of output data buffer */
		 ) {
  
  ubsec_hash_io_t  cmd;
  int              status = 0;
  
  bzero(&cmd, sizeof(cmd));

  cmd.SourceBuffer      = data_in;
  cmd.SourceBufferBytes = data_in_len;
  cmd.auth_alg          = auth_alg;
  cmd.HashBufferBytes   = data_out_len;

  status = ubsec_ioctl(fd, (unsigned char *)&cmd, UBSEC_SSL_HASH_FUNC);

  if(status == 0) {
    memcpy(data_out, cmd.HashBuffer, data_out_len);
  } else {
    bzero(data_out, data_out_len);
  }
  
  user_timing = cmd.time_us;
  
  return(status);
}









